@@ -38,10 +38,12 @@ static struct zbc_sg_sk_s {
 	enum zbc_sk	sk;
 	const char	*sk_name;
 } zbc_sg_sk_list[] = {
+	{ ZBC_SK_NOT_READY,		"Not-ready"		},
+	{ ZBC_SK_MEDIUM_ERROR,		"Medium-error"		},
 	{ ZBC_SK_ILLEGAL_REQUEST,	"Illegal-request"	},
 	{ ZBC_SK_DATA_PROTECT,		"Data-protect"		},
 	{ ZBC_SK_ABORTED_COMMAND,	"Aborted-command"	},
-	{ 0,				NULL }
+	{ 0,				NULL			}
 };
 
 /**
@@ -84,6 +86,18 @@ static struct zbc_sg_asc_ascq_s {
 		ZBC_ASC_INSUFFICIENT_ZONE_RESOURCES,
 		"Insufficient-zone-resources"
 	},
+	{
+		ZBC_ASC_READ_ERROR,
+		"Read-error"
+	},
+	{
+		ZBC_ASC_WRITE_ERROR,
+		"Write-error"
+	},
+	{
+		ZBC_ASC_FORMAT_IN_PROGRESS,
+		"Format-in-progress"
+	},
 	{
 		0,
 		NULL
@@ -326,15 +340,20 @@ int zbc_open(const char *filename, int flags, struct zbc_device **pdev)
 	struct zbc_device *dev = NULL;
 	unsigned int allowed_drv;
 	char *path = NULL;
-	int ret, i;
+	int ret, i, mask = ZBC_O_DRV_MASK;
 
 	ret = zbc_realpath(filename, &path);
 	if (ret)
 		return ret;
 
-	allowed_drv = flags & ZBC_O_DRV_MASK;
+#ifdef HAVE_DEVTEST
+	if (flags & ZBC_O_DEVTEST)
+		mask = ZBC_O_TEST_DRV_MASK;
+#endif
+
+	allowed_drv = flags & mask;
 	if (!allowed_drv)
-		allowed_drv = ZBC_O_DRV_MASK;
+		allowed_drv = mask;
 
 	/* Test all backends until one accepts the drive */
 	ret = -ENODEV;
@@ -469,8 +488,12 @@ int zbc_report_zones(struct zbc_device *dev, uint64_t sector,
 		     enum zbc_reporting_options ro,
 		     struct zbc_zone *zones, unsigned int *nr_zones)
 {
+        unsigned int max_rep_zones, max_zones = *nr_zones;
         unsigned int n, nz = 0;
 	uint64_t last_sector;
+	uint8_t *buf = NULL;
+	size_t bufsz = 0;
+	size_t max_bufsz;
 	int ret;
 
 	if (!zbc_test_mode(dev) &&
@@ -480,46 +503,78 @@ int zbc_report_zones(struct zbc_device *dev, uint64_t sector,
 		return 0;
 	}
 
+	/*
+	 * Allocate a buffer for the backend driver to do the report.
+	 * We have 2 cases: zones is NULL, in which case we only want to
+	 * count zones. For this, allocate the minimum buffer size.
+	 * Otherwise, allocate a buffer to fit nr_zones to report, up to
+	 * the maximum transfer size.
+	 */
 	if (!zones) {
-		/* Get the number of zones */
-		*nr_zones = 0;
-		return (dev->zbd_drv->zbd_report_zones)(dev, sector,
-							zbc_ro_mask(ro),
-							NULL, nr_zones);
+		bufsz = dev->zbd_report_bufsz_min;
+	} else {
+		bufsz = 64;
+		bufsz += max_zones * 64;
+		bufsz =(bufsz + dev->zbd_report_bufsz_mask) &
+			~dev->zbd_report_bufsz_mask;
+		max_bufsz = dev->zbd_info.zbd_max_rw_sectors << 9;
+		if (bufsz > max_bufsz)
+			bufsz = max_bufsz;
+		else if (bufsz < dev->zbd_report_bufsz_min)
+			bufsz = dev->zbd_report_bufsz_min;
+		max_rep_zones = (bufsz - 64) / 64;
+	}
+
+	if (posix_memalign((void **) &buf, PAGE_SIZE, bufsz) != 0) {
+		zbc_error("No memory for report zones buffer (%zu B)\n",
+			  bufsz);
+		return -ENOMEM;
+	}
+
+	if (!zones) {
+		ret = (dev->zbd_drv->zbd_report_zones)(dev, sector,
+						zbc_ro_mask(ro),
+						NULL, &nz, buf, bufsz);
+		goto out;
 	}
 
         /* Get zones information */
-        while (nz < *nr_zones) {
+	while (nz < max_zones &&
+	       sector < dev->zbd_info.zbd_sectors) {
+
+		if (nz + max_rep_zones > max_zones)
+			n = max_zones - nz;
+		else
+			n = max_rep_zones;
 
-		n = *nr_zones - nz;
 		ret = (dev->zbd_drv->zbd_report_zones)(dev, sector,
 					zbc_ro_mask(ro) | ZBC_RO_PARTIAL,
-					&zones[nz], &n);
+					&zones[nz], &n, buf, bufsz);
 		if (ret != 0) {
 			zbc_error("%s: Get zones from sector %llu failed %d (%s)\n",
 				  dev->zbd_filename,
 				  (unsigned long long) sector,
 				  ret, strerror(-ret));
-			return ret;
+			goto out;
 		}
 
 		if (n == 0)
 			break;
 
 		nz += n;
+
 		last_sector = zones[nz - 1].zbz_start +
 			zones[nz - 1].zbz_length;
-
 		if (last_sector >= dev->zbd_info.zbd_sectors)
 			break;
-
 		sector = last_sector;
-
         }
 
+out:
 	*nr_zones = nz;
+	free(buf);
 
-	return 0;
+	return ret;
 }
 
 /**
