@@ -70,6 +70,46 @@ enum zbc_ata_drv_flags {
 
 };
 
+char *zbc_ata_cmd_name(struct zbc_sg_cmd *cmd)
+{
+	switch (cmd->cdb[14]) {
+	case ZBC_ATA_IDENTIFY:
+		return "/IDENTIFY";
+	case ZBC_ATA_EXEC_DEV_DIAGNOSTIC:
+		return "/EXEC DEV DIAGNOSTIC";
+	case ZBC_ATA_READ_LOG_DMA_EXT:
+		return "/READ LOG DMA EXT";
+	case ZBC_ATA_SET_FEATURES:
+		return "/SET FEATURES";
+	case ZBC_ATA_REQUEST_SENSE_DATA_EXT:
+		return "/REQUEST SENSE DATA EXT";
+	case ZBC_ATA_READ_DMA_EXT:
+		return "/READ DMA EXT";
+	case ZBC_ATA_WRITE_DMA_EXT:
+		return "/WRITE DMA EXT";
+	case ZBC_ATA_FLUSH_CACHE_EXT:
+		return "/FLUSH CACHE EXT";
+	case ZBC_ATA_ENABLE_SENSE_DATA_REPORTING:
+		return "/ENABLE SENSE DATA REPORTING";
+	case ZBC_ATA_ZAC_MANAGEMENT_IN:
+		return "/REPORT ZONES EXT";
+	case ZBC_ATA_ZAC_MANAGEMENT_OUT:
+		switch (cmd->cdb[4]) {
+		case ZBC_ATA_CLOSE_ZONE_EXT_AF:
+			return "/CLOSE ZONE EXT";
+		case ZBC_ATA_FINISH_ZONE_EXT_AF:
+			return "/FINISH ZONE EXT";
+		case ZBC_ATA_OPEN_ZONE_EXT_AF:
+			return "/OPEN ZONE EXT";
+		case ZBC_ATA_RESET_WRITE_POINTER_EXT_AF:
+			return "/RESET WRITE POINTER EXT";
+		}
+		break;
+	}
+
+	return "/UNKNOWN COMMAND";
+}
+
 /**
  * Get a word from a command data buffer.
  */
@@ -751,6 +791,7 @@ static int zbc_ata_native_flush(struct zbc_device *dev)
 
 	/* Fill command CDB */
 	cmd.io_hdr.dxfer_direction = SG_DXFER_NONE;
+	cmd.io_hdr.timeout *= 2;
 	cmd.cdb[0] = ZBC_SG_ATA16_CDB_OPCODE;
 	/* Non-Data protocol, ext=1 */
 	cmd.cdb[1] = (0x3 << 1) | 0x01;
@@ -781,26 +822,16 @@ static int zbc_ata_flush(struct zbc_device *dev)
  */
 static int zbc_ata_report_zones(struct zbc_device *dev, uint64_t sector,
 				enum zbc_reporting_options ro,
-				struct zbc_zone *zones, unsigned int *nr_zones)
+				struct zbc_zone *zones, unsigned int *nr_zones,
+				uint8_t *buf, size_t bufsz)
 {
-	size_t bufsz = ZBC_ZONE_DESCRIPTOR_OFFSET;
 	uint64_t lba = zbc_dev_sect2lba(dev, sector);
 	unsigned int i, nz = 0, buf_nz;
-	size_t max_bufsz;
 	struct zbc_sg_cmd cmd;
-	uint8_t *buf;
 	int ret;
 
-	if (*nr_zones)
-		bufsz += (size_t)*nr_zones * ZBC_ZONE_DESCRIPTOR_LENGTH;
-
-	bufsz = (bufsz + 4095) & ~4095;
-	max_bufsz = dev->zbd_info.zbd_max_rw_sectors << 9;
-	if (bufsz > max_bufsz)
-		bufsz = max_bufsz;
-
 	/* Allocate and intialize report zones command */
-	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_ATA16, NULL, bufsz);
+	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_ATA16, buf, bufsz);
 	if (ret != 0)
 		return ret;
 
@@ -880,7 +911,6 @@ static int zbc_ata_report_zones(struct zbc_device *dev, uint64_t sector,
 	}
 
 	/* Get number of zones in result */
-	buf = (uint8_t *) cmd.buf;
 	nz = zbc_ata_get_dword(buf) / ZBC_ZONE_DESCRIPTOR_LENGTH;
 	/* max_lba = zbc_ata_get_qword(&buf[8]); */
 
@@ -1217,7 +1247,7 @@ out:
 /**
  * Get a device capacity information (total blocks & block size).
  */
-static int zbc_ata_get_capacity(struct zbc_device *dev)
+int zbc_ata_get_capacity(struct zbc_device *dev)
 {
 	uint8_t buf[512];
 	uint64_t qword;
@@ -1231,7 +1261,7 @@ static int zbc_ata_get_capacity(struct zbc_device *dev)
 			       buf,
 			       sizeof(buf));
 	if (ret != 0) {
-		zbc_error("%s: Get supported capabilities page failed\n",
+		zbc_error("%s: Get capacity page failed\n",
 			  dev->zbd_filename);
 		return ret;
 	}
@@ -1277,9 +1307,6 @@ static int zbc_ata_get_capacity(struct zbc_device *dev)
 		return -EINVAL;
 	}
 
-	/* Get maximum command size */
-	zbc_sg_get_max_cmd_blocks(dev);
-
 	dev->zbd_info.zbd_pblock_size =
 		dev->zbd_info.zbd_lblock_size * logical_per_physical;
 	dev->zbd_info.zbd_pblocks =
@@ -1314,6 +1341,10 @@ static int zbc_ata_get_dev_info(struct zbc_device *dev)
 {
 	int ret;
 
+	/* Always use 512B aligned zone report buffers */
+	dev->zbd_report_bufsz_min = 512;
+	dev->zbd_report_bufsz_mask = dev->zbd_report_bufsz_min - 1;
+
 	/* Make sure the device is ready */
 	ret = zbc_sg_test_unit_ready(dev);
 	if (ret != 0)
@@ -1337,6 +1368,9 @@ static int zbc_ata_get_dev_info(struct zbc_device *dev)
 	if (ret != 0)
 		return ret;
 
+	/* Get maximum command size */
+	zbc_sg_get_max_cmd_blocks(dev);
+
 	/* Check if we have a functional SAT for read/write */
 	if (!zbc_test_mode(dev))
 		zbc_ata_test_sbc_sat(dev);
@@ -1456,7 +1490,7 @@ static int zbc_ata_open(const char *filename,
 
 	*pdev = dev;
 
-	zbc_debug("%s: ########## ATA driver succeeded ##########\n",
+	zbc_debug("%s: ########## ATA driver succeeded ##########\n\n",
 		  filename);
 
 	return 0;
@@ -1471,7 +1505,7 @@ out:
 	if (fd >= 0)
 		close(fd);
 
-	zbc_debug("%s: ########## ATA driver failed %d ##########\n",
+	zbc_debug("%s: ########## ATA driver failed %d ##########\n\n",
 		  filename, ret);
 
 	return ret;
