@@ -9,6 +9,7 @@
  * Author: Damien Le Moal (damien.lemoal@wdc.com)
  */
 #include "zbc.h"
+#include "zbc_utils.h"
 #include "zbc_sg.h"
 
 #include <assert.h>
@@ -68,8 +69,12 @@ static int dir_has(const char *dir, const char *entry)
 	DIR *d;
 
 	d = opendir(dir);
-	if (!d)
+	if (!d) {
+		/* If dir does not exist, entry does not exist either */
+		if (errno == ENOENT)
+			return 0;
 		return -errno;
+	}
 	while ((e = readdir(d)) != NULL) {
 		if (strcmp(e->d_name, entry) == 0) {
 			res = 1;
@@ -114,32 +119,20 @@ out:
 static int zbc_block_get_partition_start(struct zbc_device *dev)
 {
 	struct zbc_block_device *zbd = zbc_dev_to_block(dev);
-	char str[128];
-	FILE *file;
+	char sysfs_path[128];
 	int ret;
 
 	/* Open the start offset file of the partition */
-	snprintf(str, sizeof(str),
+	snprintf(sysfs_path, sizeof(sysfs_path),
 		 "/sys/block/%s/%s/start",
 		 zbd->holder_name,
 		 zbd->part_name);
-	file = fopen(str, "r");
-	if (!file) {
-		int ret = -errno;
-
-		zbc_error("%s: open %s failed %d (%s)\n",
-			  zbd->part_name, str,
-			  errno, strerror(errno));
-		return ret;
-	}
-
-	ret = fscanf(file, "%llu", &zbd->part_offset);
-	if (ret <= 0) {
+	ret = zbc_get_sysfs_val_ull(sysfs_path, &zbd->part_offset);
+	if (ret) {
 		zbc_error("%s: can't read partition offset from %s\n",
-			  zbd->part_name, str);
+			  zbd->part_name, sysfs_path);
 		return ret;
 	}
-	fclose(file);
 
 	zbc_debug("%s: Partition of %s, start sector offset %llu\n",
 		  dev->zbd_filename,
@@ -243,41 +236,29 @@ out:
 static int zbc_block_device_classify(struct zbc_device *dev)
 {
 	struct zbc_block_device *zbd = zbc_dev_to_block(dev);
-	char str[128];
-	FILE *file;
+	char model[32];
 	int ret;
 
 	/* Check that this is a zoned block device */
-	snprintf(str, sizeof(str),
-		 "/sys/block/%s/queue/zoned",
-		 zbd->holder_name);
-	file = fopen(str, "r");
-	if (!file)
+	ret = zbc_get_sysfs_queue_str(zbd->holder_name, "zoned",
+				      model, sizeof(model));
+	if (ret)
 		/*
 		 * Cannot determine type: go on with SCSI,
 		 * ATA or fake backends.
 		 */
 		return -ENXIO;
 
-	memset(str, 0, sizeof(str));
-	ret = fscanf(file, "%s", str);
-	if (ret <= 0) {
-		zbc_error("%s: can't read zoned model from %s\n",
-			  zbd->part_name, str);
-		return ret;
-	}
-	fclose(file);
-
-	if (strcmp(str, "host-aware") == 0) {
+	if (strcmp(model, "host-aware") == 0) {
 		dev->zbd_info.zbd_model = ZBC_DM_HOST_AWARE;
-	} else if (strcmp(str, "host-managed") == 0) {
+	} else if (strcmp(model, "host-managed") == 0) {
 		dev->zbd_info.zbd_model = ZBC_DM_HOST_MANAGED;
-	} else if (strcmp(str, "none") == 0) {
+	} else if (strcmp(model, "none") == 0) {
 		dev->zbd_info.zbd_model = ZBC_DM_STANDARD;
 		return -ENXIO;
 	} else {
 		zbc_debug("%s: Unknown device model \"%s\"\n",
-			  dev->zbd_filename, str);
+			  dev->zbd_filename, model);
 		dev->zbd_info.zbd_model = ZBC_DM_DRIVE_UNKNOWN;
 		return -ENXIO;
 	}
@@ -285,32 +266,6 @@ static int zbc_block_device_classify(struct zbc_device *dev)
 	return 0;
 }
 
-/**
- * Get a string in a file and strip it of trailing
- * spaces and carriage return.
- */
-static int zbc_block_get_str(FILE *file, char *str)
-{
-	int len = 0;
-
-	if (fgets(str, 128, file)) {
-		len = strlen(str) - 1;
-		while (len > 0) {
-			if (str[len] == ' ' ||
-			    str[len] == '\t' ||
-			    str[len] == '\r' ||
-			    str[len] == '\n') {
-				str[len] = '\0';
-				len--;
-			} else {
-				break;
-			}
-		}
-	}
-
-	return len;
-}
-
 /**
  * Get vendor ID.
  */
@@ -318,47 +273,28 @@ static int zbc_block_get_vendor_id(struct zbc_device *dev)
 {
 	struct zbc_block_device *zbd = zbc_dev_to_block(dev);
 	char str[128];
-	FILE *file;
-	int n = 0, len;
+	int ret, n = 0;
 
-	snprintf(str, sizeof(str),
-		 "/sys/block/%s/device/vendor",
-		 zbd->holder_name);
-	file = fopen(str, "r");
-	if (file) {
-		len = zbc_block_get_str(file, str);
-		if (len)
-			n = snprintf(dev->zbd_info.zbd_vendor_id,
-				     ZBC_DEVICE_INFO_LENGTH,
-				     "%s ", str);
-		fclose(file);
-	}
+	ret = zbc_get_sysfs_device_str(zbd->holder_name, "vendor",
+				       str, sizeof(str));
+	if (!ret)
+		n = snprintf(dev->zbd_info.zbd_vendor_id,
+			     ZBC_DEVICE_INFO_LENGTH,
+			     "%s ", str);
 
-	snprintf(str, sizeof(str),
-		 "/sys/block/%s/device/model",
-		 zbd->holder_name);
-	file = fopen(str, "r");
-	if (file) {
-		len = zbc_block_get_str(file, str);
-		if (len)
-			n += snprintf(&dev->zbd_info.zbd_vendor_id[n],
-				      ZBC_DEVICE_INFO_LENGTH - n,
-				      "%s ", str);
-		fclose(file);
-	}
+	ret = zbc_get_sysfs_device_str(zbd->holder_name, "model",
+				       str, sizeof(str));
+	if (!ret)
+		n += snprintf(&dev->zbd_info.zbd_vendor_id[n],
+			      ZBC_DEVICE_INFO_LENGTH - n,
+			      "%s ", str);
 
-	snprintf(str, sizeof(str),
-		 "/sys/block/%s/device/rev",
-		 zbd->holder_name);
-	file = fopen(str, "r");
-	if (file) {
-		len = zbc_block_get_str(file, str);
-		if (len)
-			n += snprintf(&dev->zbd_info.zbd_vendor_id[n],
-				      ZBC_DEVICE_INFO_LENGTH - n,
-				      "%s", str);
-		fclose(file);
-	}
+	ret = zbc_get_sysfs_device_str(zbd->holder_name, "rev",
+				       str, sizeof(str));
+	if (!ret)
+		n += snprintf(&dev->zbd_info.zbd_vendor_id[n],
+			      ZBC_DEVICE_INFO_LENGTH - n,
+			      "%s", str);
 
 	return n > 0;
 }
@@ -369,38 +305,41 @@ static int zbc_block_get_vendor_id(struct zbc_device *dev)
 static int zbc_block_get_zone_sectors(struct zbc_device *dev)
 {
 	struct zbc_block_device *zbd = zbc_dev_to_block(dev);
-	char str[128];
-	FILE *file;
 	int ret;
 
 	/* Open the chunk_sectors file */
-	snprintf(str, sizeof(str),
-		 "/sys/block/%s/queue/chunk_sectors",
-		 zbd->holder_name);
-	file = fopen(str, "r");
-	if (!file) {
-		int ret = -errno;
-
-		zbc_error("%s: open %s failed %d (%s)\n",
-			  zbd->part_name, str,
-			  errno, strerror(errno));
+	ret = zbc_get_sysfs_queue_val_ull(zbd->holder_name, "chunk_sectors",
+					  &zbd->zone_sectors);
+	if (ret) {
+		zbc_error("%s: get zone sectors from sysfs failed\n",
+			  zbd->part_name);
 		return ret;
 	}
 
-	ret = fscanf(file, "%llu", &zbd->zone_sectors);
-	if (ret <= 0) {
-		zbc_error("%s: can't read zone sectors from %s\n",
-			  zbd->part_name, str);
-		return ret;
-	}
-	fclose(file);
-
 	zbc_debug("%s: Zones of %llu sectors\n",
 		  zbd->part_name, zbd->zone_sectors);
 
 	return 0;
 }
 
+/**
+ * Get the maximum number of open zones of a device.
+ */
+static unsigned int zbc_block_get_max_open_zones(struct zbc_device *dev)
+{
+	struct zbc_block_device *zbd = zbc_dev_to_block(dev);
+	unsigned long long val;
+	int ret;
+
+	/* Open the max_open_zones file */
+	ret = zbc_get_sysfs_queue_val_ull(zbd->holder_name,
+					  "max_open_zones", &val);
+	if (ret)
+		return ZBC_BLOCK_MAX_OPEN_ZONES;
+
+	return val;
+}
+
 /**
  * Test if the device can be handled and get the block device info.
  */
@@ -411,6 +350,14 @@ static int zbc_block_get_info(struct zbc_device *dev, struct stat *st)
 	int size32;
 	int ret;
 
+	/*
+	 * We do not need zbc_report_zones() to allocate a buffer for
+	 * report zones. zbc_block_report_zones() does its own allocation
+	 * as needed.
+	 */
+	dev->zbd_report_bufsz_min = 0;
+	dev->zbd_report_bufsz_mask = 0;
+
 	/* Check if we are dealing with a partition */
 	ret = zbc_block_handle_partition(dev);
 	if (ret)
@@ -504,7 +451,7 @@ static int zbc_block_get_info(struct zbc_device *dev, struct stat *st)
 		dev->zbd_info.zbd_flags |= ZBC_UNRESTRICTED_READ;
 		if (dev->zbd_info.zbd_model == ZBC_DM_HOST_MANAGED) {
 			dev->zbd_info.zbd_max_nr_open_seq_req =
-				ZBC_BLOCK_MAX_OPEN_ZONES;
+				zbc_block_get_max_open_zones(dev);
 			dev->zbd_info.zbd_opt_nr_open_seq_pref = 0;
 			dev->zbd_info.zbd_opt_nr_non_seq_write_seq_pref = 0;
 		} else {
@@ -537,7 +484,7 @@ static int zbc_block_open(const char *filename, int flags,
 	struct zbc_block_device *zbd;
 	struct zbc_device *dev;
 	struct stat st;
-	int fd, ret;
+	int fd = -1, ret;
 
 	zbc_debug("%s: ########## Trying BLOCK driver ##########\n",
 		  filename);
@@ -551,8 +498,10 @@ static int zbc_block_open(const char *filename, int flags,
 		return ret;
 	}
 
-	if (!S_ISBLK(st.st_mode))
-		return -ENXIO;
+	if (!S_ISBLK(st.st_mode)) {
+		ret = -ENXIO;
+		goto out;
+	}
 
 	/* Open block device */
 	fd = open(filename, (flags & ZBC_O_DMODE_MASK) | O_LARGEFILE);
@@ -584,7 +533,7 @@ static int zbc_block_open(const char *filename, int flags,
 
 	*pdev = dev;
 
-	zbc_debug("%s: ########## BLOCK driver succeeded ##########\n",
+	zbc_debug("%s: ########## BLOCK driver succeeded ##########\n\n",
 		  filename);
 
 	return 0;
@@ -602,7 +551,7 @@ out:
 	if (fd >= 0)
 		close(fd);
 
-	zbc_debug("%s: ########## BLOCK driver failed %d ##########\n",
+	zbc_debug("%s: ########## BLOCK driver failed %d ##########\n\n",
 		  filename, ret);
 
 	return ret;
@@ -677,9 +626,9 @@ static bool zbc_block_must_report(struct zbc_zone *zone, uint64_t start_sector,
  * Get the block device zone information.
  */
 static int zbc_block_report_zones(struct zbc_device *dev, uint64_t start_sector,
-				  enum zbc_reporting_options ro,
-				  struct zbc_zone *zones,
-				  unsigned int *nr_zones)
+				enum zbc_reporting_options ro,
+				struct zbc_zone *zones, unsigned int *nr_zones,
+				uint8_t *buf, size_t bufsz)
 {
 	size_t rep_size;
 	uint64_t sector = start_sector;
@@ -772,7 +721,7 @@ static int zbc_block_reset_one(struct zbc_device *dev, uint64_t sector)
 
 	/* Get zone info */
 	ret = zbc_block_report_zones(dev, sector, ZBC_RO_ALL,
-				     &zone, &nr_zones);
+				     &zone, &nr_zones, NULL, 0);
 	if (ret)
 		return ret;
 
@@ -783,11 +732,6 @@ static int zbc_block_reset_one(struct zbc_device *dev, uint64_t sector)
 		return -EINVAL;
 	}
 
-	if (zbc_zone_conventional(&zone)
-	    || zbc_zone_empty(&zone))
-		/* Nothing to do */
-		return 0;
-
 	/* Reset zone */
 	range.sector = zbc_zone_start(&zone);
 	range.nr_sectors = zbc_zone_length(&zone);
@@ -854,7 +798,7 @@ static int zbc_block_reset_all(struct zbc_device *dev)
 		/* Get zone info */
 		nr_zones = ZBC_BLOCK_ZONE_REPORT_NR_ZONES;
 		ret = zbc_block_report_zones(dev, sector, ZBC_RO_ALL,
-					     zones, &nr_zones);
+					     zones, &nr_zones, NULL, 0);
 		if (ret || !nr_zones)
 			break;
 
@@ -1073,9 +1017,9 @@ static int zbc_block_close(struct zbc_device *dev)
 }
 
 static int zbc_block_report_zones(struct zbc_device *dev, uint64_t sector,
-				  enum zbc_reporting_options ro,
-				  struct zbc_zone *zones,
-				  unsigned int *nr_zones)
+				enum zbc_reporting_options ro,
+				struct zbc_zone *zones, unsigned int *nr_zones,
+				uint8_t *buf, size_t bufsz)
 {
 	return -EOPNOTSUPP;
 }
