@@ -406,6 +406,10 @@ static int zbc_fake_set_info(struct zbc_device *dev)
 	/* Get maximum command size */
 	zbc_sg_get_max_cmd_blocks(dev);
 
+	/* We do not need a buffer for report zones */
+	dev->zbd_report_bufsz_min = 0;
+	dev->zbd_report_bufsz_mask = 0;
+
 	return 0;
 }
 
@@ -459,7 +463,7 @@ static int zbc_fake_open(const char *filename, int flags,
 
 	*pdev = &fdev->dev;
 
-	zbc_debug("%s: ########## FAKE driver succeeded ##########\n",
+	zbc_debug("%s: ########## FAKE driver succeeded ##########\n\n",
 		  filename);
 
 	return 0;
@@ -473,7 +477,7 @@ out_free_dev:
 out:
 	close(fd);
 
-	zbc_debug("%s: ########## FAKE driver failed %d ##########\n",
+	zbc_debug("%s: ########## FAKE driver failed %d ##########\n\n",
 		  filename, ret);
 
 	return ret;
@@ -544,7 +548,8 @@ static bool zbc_fake_must_report_zone(struct zbc_zone *zone,
  */
 static int zbc_fake_report_zones(struct zbc_device *dev, uint64_t sector,
 				 enum zbc_reporting_options ro,
-				 struct zbc_zone *zones, unsigned int *nr_zones)
+				 struct zbc_zone *zones, unsigned int *nr_zones,
+				 uint8_t *buf, size_t bufsz)
 {
 	struct zbc_fake_device *fdev = zbc_fake_to_file_dev(dev);
 	unsigned int max_nr_zones = *nr_zones;
@@ -1421,7 +1426,7 @@ static int zbc_fake_set_write_pointer(struct zbc_device *dev,
 {
 	struct zbc_fake_device *fdev = zbc_fake_to_file_dev(dev);
 	struct zbc_zone *zone;
-	int ret = -EIO;
+	int ret = -EINVAL;
 
 	if (!fdev->zbd_meta) {
 		zbc_set_errno(ZBC_SK_NOT_READY, ZBC_ASC_FORMAT_IN_PROGRESS);
@@ -1434,24 +1439,24 @@ static int zbc_fake_set_write_pointer(struct zbc_device *dev,
 	if (!zone)
 		goto out;
 
-	/* Do nothing for conventional zones */
-	if (zbc_zone_sequential_req(zone)) {
+	if (zbc_zone_conventional(zone))
+		goto out;
 
-		if (zbc_zone_is_open(zone))
-			zbc_zone_do_close(fdev, zone);
+	if (wp_sector < zone->zbz_start ||
+	    wp_sector > zone->zbz_start + zone->zbz_length)
+		goto out;
 
-		zone->zbz_write_pointer = wp_sector;
-		if (zone->zbz_write_pointer == zone->zbz_start) {
-			zone->zbz_condition = ZBC_ZC_EMPTY;
-		} else if (zone->zbz_write_pointer > zone->zbz_start &&
-			   zone->zbz_write_pointer <
-			   zone->zbz_start + zone->zbz_length) {
-			zone->zbz_condition = ZBC_ZC_CLOSED;
-		} else {
-			zone->zbz_condition = ZBC_ZC_FULL;
-			zone->zbz_write_pointer = (uint64_t)-1;
-		}
+	if (zbc_zone_is_open(zone))
+		zbc_zone_do_close(fdev, zone);
 
+	zone->zbz_write_pointer = wp_sector;
+	if (zone->zbz_write_pointer == zone->zbz_start) {
+		zone->zbz_condition = ZBC_ZC_EMPTY;
+	} else if (zone->zbz_write_pointer == zone->zbz_start + zone->zbz_length) {
+		zone->zbz_condition = ZBC_ZC_FULL;
+		zone->zbz_write_pointer = (uint64_t)-1;
+	} else {
+		zone->zbz_condition = ZBC_ZC_CLOSED;
 	}
 
 	ret = 0;
