@@ -34,50 +34,37 @@
  * SCSI commands reply length.
  */
 #define ZBC_SCSI_INQUIRY_BUF_LEN	96
+#define ZBC_SCSI_VPD_PAGE_00_LEN	32
 #define ZBC_SCSI_VPD_PAGE_B1_LEN	64
 #define ZBC_SCSI_VPD_PAGE_B6_LEN	64
 #define ZBC_SCSI_READ_CAPACITY_BUF_LEN	32
 
+/**
+ * Driver device flags.
+ */
+enum zbc_scsi_drv_flags {
+	/** This is an ATA drive used through a SAT */
+	ZBC_IS_ATA	= 0x00000001,
+};
+
 /**
  * Fill the buffer with the result of INQUIRY command.
  * @buf must be at least ZBC_SG_INQUIRY_REPLY_LEN bytes long.
  */
 static int zbc_scsi_inquiry(struct zbc_device *dev,
-			    uint8_t page,
-			    void *buf,
-			    uint16_t buf_len)
+			    void *buf, uint16_t buf_len)
 {
 	struct zbc_sg_cmd cmd;
 	int ret;
 
 	/* Allocate and intialize inquiry command */
+	memset(buf, 0, buf_len);
 	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_INQUIRY, buf, buf_len);
 	if (ret != 0)
 		return ret;
 
-	/* Fill command CDB:
-	 * +=============================================================================+
-	 * |  Bit|   7    |   6    |   5    |   4    |   3    |   2    |   1    |   0    |
-	 * |Byte |        |        |        |        |        |        |        |        |
-	 * |=====+=======================================================================|
-	 * | 0   |                           Operation Code (12h)                        |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 1   | Logical Unit Number      |                  Reserved         |  EVPD  |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 2   |                           Page Code                                   |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 3   | (MSB)                                                                 |
-	 * |- - -+---                    Allocation Length                            ---|
-	 * | 4   |                                                                 (LSB) |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 5   |                           Control                                     |
-	 * +=============================================================================+
-	 */
+	/* Fill command CDB */
 	cmd.cdb[0] = ZBC_SG_INQUIRY_CDB_OPCODE;
-	if (page) {
-		cmd.cdb[1] = 0x01;
-		cmd.cdb[2] = page;
-	}
 	zbc_sg_set_int16(&cmd.cdb[3], buf_len);
 
 	/* Execute the SG_IO command */
@@ -89,62 +76,67 @@ static int zbc_scsi_inquiry(struct zbc_device *dev,
 }
 
 /**
- * Test is ZBC/ZAC SAT is working.
+ * Fill the buffer with the result of a VPD page INQUIRY command.
  */
-static int zbc_scsi_test_sat(struct zbc_device *dev)
+static int zbc_scsi_vpd_inquiry(struct zbc_device *dev,
+				uint8_t page,
+				void *buf, uint16_t buf_len)
 {
-	size_t bufsz = 512;
 	struct zbc_sg_cmd cmd;
 	int ret;
 
-	/* Allocate and intialize report zones command */
-	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_REPORT_ZONES, NULL, bufsz);
+	/* Allocate and intialize inquiry command */
+	memset(buf, 0, buf_len);
+	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_INQUIRY, buf, buf_len);
 	if (ret != 0)
 		return ret;
 
-	/* Fill command CDB:
-	 * +=============================================================================+
-	 * |  Bit|   7    |   6    |   5    |   4    |   3    |   2    |   1    |   0    |
-	 * |Byte |        |        |        |        |        |        |        |        |
-	 * |=====+==========================+============================================|
-	 * | 0   |                           Operation Code (95h)                        |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 1   |      Reserved            |       Service Action (00h)                 |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 2   | (MSB)                                                                 |
-	 * |- - -+---                        Zone Start LBA                           ---|
-	 * | 9   |                                                                 (LSB) |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 10  | (MSB)                                                                 |
-	 * |- - -+---                        Allocation Length                        ---|
-	 * | 13  |                                                                 (LSB) |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 14  |Partial |Reserved|                 Reporting Options                   |
-	 * |-----+-----------------------------------------------------------------------|
-	 * | 15  |                           Control                                     |
-	 * +=============================================================================+
-	 */
-	cmd.cdb[0] = ZBC_SG_REPORT_ZONES_CDB_OPCODE;
-	cmd.cdb[1] = ZBC_SG_REPORT_ZONES_CDB_SA;
-	zbc_sg_set_int64(&cmd.cdb[2], 0);
-	zbc_sg_set_int32(&cmd.cdb[10], (unsigned int) bufsz);
-	cmd.cdb[14] = ZBC_RO_PARTIAL;
+	/* Fill command CDB */
+	cmd.cdb[0] = ZBC_SG_INQUIRY_CDB_OPCODE;
+	cmd.cdb[1] = 0x01;
+	cmd.cdb[2] = page;
+	zbc_sg_set_int16(&cmd.cdb[3], buf_len);
 
-	/* Send the SG_IO command */
+	/* Execute the SG_IO command */
 	ret = zbc_sg_cmd_exec(dev, &cmd);
 
 	zbc_sg_cmd_destroy(&cmd);
 
-	if (ret != 0)
-		zbc_debug("%s: ZBC/ZAC SAT not supported\n",
-			  dev->zbd_filename);
-	else
-		zbc_debug("%s: ZBC/ZAC SAT supported: treating ATA device as SCSI\n",
-			  dev->zbd_filename);
-
 	return ret;
 }
 
+/**
+ * Test if a VPD page is supported.
+ */
+static bool zbc_scsi_vpd_page_supported(struct zbc_device *dev, uint8_t page)
+{
+	uint8_t buf[ZBC_SCSI_VPD_PAGE_00_LEN];
+	int vpd_len, i, ret;
+
+	ret = zbc_scsi_vpd_inquiry(dev, 0x00, buf, ZBC_SCSI_VPD_PAGE_00_LEN);
+	if (ret != 0) {
+		zbc_error("%s: zbc_scsi_vpd_inquiry VPD page 0x00 failed\n",
+			  dev->zbd_filename);
+		return false;
+	}
+
+	if (buf[1] != 0x00) {
+		zbc_error("%s: Invalid page code 0x%02x for VPD page 0x00\n",
+			  dev->zbd_filename, (int)buf[1]);
+		return false;
+	}
+
+	vpd_len = zbc_sg_get_int16(&buf[2]) + 4;
+	if (vpd_len > ZBC_SCSI_VPD_PAGE_00_LEN)
+		vpd_len = ZBC_SCSI_VPD_PAGE_00_LEN;
+	for (i = 4; i < vpd_len; i++) {
+		if (buf[i] == page)
+			return true;
+	}
+
+	return false;
+}
+
 /**
  * Get information string from inquiry output.
  */
@@ -182,31 +174,29 @@ static int zbc_scsi_classify(struct zbc_device *dev)
 	int ret;
 
 	/* Get device info */
-	ret = zbc_scsi_inquiry(dev, 0, buf, ZBC_SCSI_INQUIRY_BUF_LEN);
+	ret = zbc_scsi_inquiry(dev, buf, ZBC_SCSI_INQUIRY_BUF_LEN);
 	if (ret != 0) {
 		zbc_error("%s: zbc_scsi_inquiry failed\n",
 			  dev->zbd_filename);
 		return ret;
 	}
 
-	/*
-	 * If this is an ATA drive, try to see if SAT is working.
-	 * If SAT is working, treat the disk as SCSI.
-	 */
-	if (strncmp((char *)&buf[8], "ATA", 3) == 0) {
-		ret = zbc_scsi_test_sat(dev);
-		if (ret != 0)
-			return -ENXIO;
-	}
-
 	/* This is a SCSI device */
 	dev->zbd_info.zbd_type = ZBC_DT_SCSI;
 
+	/*
+	 * Check if we are dealing with an ATA device by checking for the
+	 * ATA Information VPD page (89h).
+	 */
+	if (zbc_scsi_vpd_page_supported(dev, 0x89)) {
+		dev->zbd_drv_flags |= ZBC_IS_ATA;
+		zbc_debug("%s: ATA device\n", dev->zbd_filename);
+	}
+
 	/*
 	 * Concatenate vendor identification, product identification
 	 * and product revision strings.
 	 */
-	//memset(dev->zbd_info.zbd_vendor_id, 0, ZBC_DEVICE_INFO_LENGTH);
 	memset(vid, 0, sizeof(vid));
 	memset(pid, 0, sizeof(pid));
 	memset(rev, 0, sizeof(rev));
@@ -244,10 +234,9 @@ static int zbc_scsi_classify(struct zbc_device *dev)
 	 * VPD page (B1h) to be sure. Also check that no weird value is
 	 * reported by the zoned field for host-managed devices.
 	 */
-	memset(buf, 0, sizeof(buf));
-	ret = zbc_scsi_inquiry(dev, 0xB1, buf, ZBC_SCSI_VPD_PAGE_B1_LEN);
+	ret = zbc_scsi_vpd_inquiry(dev, 0xB1, buf, ZBC_SCSI_VPD_PAGE_B1_LEN);
 	if (ret != 0) {
-		zbc_error("%s: zbc_scsi_inquiry VPD page 0xB1 failed\n",
+		zbc_error("%s: zbc_scsi_vpd_inquiry VPD page 0xB1 failed\n",
 			  dev->zbd_filename);
 		return ret;
 	}
@@ -255,7 +244,7 @@ static int zbc_scsi_classify(struct zbc_device *dev)
 	if ((buf[1] != 0xB1) ||
 	    (buf[2] != 0x00) ||
 	    (buf[3] != 0x3C)) {
-		zbc_error("%s: Invalid zbc_scsi_inquiry VPD page 0xB1 result\n",
+		zbc_error("%s: Invalid VPD page 0xB1\n",
 			  dev->zbd_filename);
 		return -EIO;
 	}
@@ -265,12 +254,14 @@ static int zbc_scsi_classify(struct zbc_device *dev)
 		if (zbc_test_mode(dev) && zoned != 0) {
 			zbc_error("%s: Invalid host-managed device ZONED field 0x%02x\n",
 				  dev->zbd_filename, zoned);
+			if (dev->zbd_drv_flags & ZBC_IS_ATA)
+				return -ENXIO;
 			return -EIO;
 		}
 		if (zoned != 0)
 			zbc_warning("%s: Invalid host-managed device ZONED field 0x%02x\n",
 				    dev->zbd_filename, zoned);
-		return 0;
+		goto out;
 	}
 
 	switch (zoned) {
@@ -302,6 +293,20 @@ static int zbc_scsi_classify(struct zbc_device *dev)
 		return -EIO;
 	}
 
+out:
+	/*
+	 * At this point, we know we have a zoned device. Check if we have the
+	 * zoned device characteristics page to check that the system SAT is
+	 * working correctly.
+	 */
+	if (!zbc_scsi_vpd_page_supported(dev, 0xB6)) {
+		zbc_error("%s: VPD page 0xb6 is not supported\n",
+			  dev->zbd_filename);
+		if (dev->zbd_drv_flags & ZBC_IS_ATA)
+			return -ENXIO;
+		return -EIO;
+	}
+
 	return 0;
 }
 
@@ -309,30 +314,18 @@ static int zbc_scsi_classify(struct zbc_device *dev)
  * Get a SCSI device zone information.
  */
 static int zbc_scsi_do_report_zones(struct zbc_device *dev, uint64_t sector,
-				    enum zbc_reporting_options ro,
-				    uint64_t *max_lba,
-				    struct zbc_zone *zones,
-				    unsigned int *nr_zones)
+				enum zbc_reporting_options ro,
+				uint64_t *max_lba,
+				struct zbc_zone *zones, unsigned int *nr_zones,
+				uint8_t *buf, size_t bufsz)
 {
-	size_t bufsz = ZBC_ZONE_DESCRIPTOR_OFFSET;
 	uint64_t lba = zbc_dev_sect2lba(dev, sector);
 	unsigned int i, nz = 0, buf_nz;
 	struct zbc_sg_cmd cmd;
-	size_t max_bufsz;
-	uint8_t *buf;
 	int ret;
 
-	if (*nr_zones)
-		bufsz += (size_t)*nr_zones * ZBC_ZONE_DESCRIPTOR_LENGTH;
-
-	/* For in kernel ATA translation: align to 512 B */
-	bufsz = (bufsz + 511) & ~511;
-	max_bufsz = dev->zbd_info.zbd_max_rw_sectors << 9;
-	if (bufsz > max_bufsz)
-		bufsz = max_bufsz;
-
-	/* Allocate and intialize report zones command */
-	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_REPORT_ZONES, NULL, bufsz);
+	/* Intialize report zones command */
+	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_REPORT_ZONES, buf, bufsz);
 	if (ret != 0)
 		return ret;
 
@@ -466,6 +459,7 @@ static int zbc_scsi_do_report_zones(struct zbc_device *dev, uint64_t sector,
 	 * | 63  |                                                                       |
 	 * +=============================================================================+
 	 */
+
 	buf += ZBC_ZONE_DESCRIPTOR_OFFSET;
 	for (i = 0; i < nz; i++) {
 
@@ -503,10 +497,11 @@ out:
  */
 static int zbc_scsi_report_zones(struct zbc_device *dev, uint64_t sector,
 				 enum zbc_reporting_options ro,
-				 struct zbc_zone *zones, unsigned int *nr_zones)
+				 struct zbc_zone *zones, unsigned int *nr_zones,
+				 uint8_t *buf, size_t bufsz)
 {
-	return zbc_scsi_do_report_zones(dev, sector, ro,
-					NULL, zones, nr_zones);
+	return zbc_scsi_do_report_zones(dev, sector, ro, NULL, zones, nr_zones,
+					buf, bufsz);
 }
 
 /**
@@ -594,6 +589,7 @@ int zbc_scsi_zone_op(struct zbc_device *dev, uint64_t sector,
 	return ret;
 }
 
+
 /**
  * Get a device capacity information (total sectors & sector sizes).
  */
@@ -606,6 +602,19 @@ static int zbc_scsi_get_capacity(struct zbc_device *dev)
 	uint64_t max_lba;
 	int ret;
 
+	/*
+	 * Some SAS HBAs have a very slow processing of the READ CAPACITY
+	 * command for ZAC ATA drives. So instead on relying on the HBA SATL
+	 * for the command translation, use the ata backend read capacity
+	 * function to directly read the data log capacity page of the disk.
+	 * In case of failure, fall back to using the READ CAPACITY command.
+	 */
+	if ((dev->zbd_drv_flags & ZBC_IS_ATA) && !zbc_test_mode(dev)) {
+		ret = zbc_ata_get_capacity(dev);
+		if (ret == 0)
+			return 0;
+	}
+
 	/* READ CAPACITY 16 */
 	ret = zbc_sg_cmd_init(dev, &cmd, ZBC_SG_READ_CAPACITY,
 			      NULL, ZBC_SCSI_READ_CAPACITY_BUF_LEN);
@@ -634,14 +643,9 @@ static int zbc_scsi_get_capacity(struct zbc_device *dev)
 	logical_per_physical = 1 << cmd.buf[13] & 0x0f;
 	max_lba = zbc_sg_get_int64(&cmd.buf[0]);
 
-	/* Get maximum command size */
-	zbc_sg_get_max_cmd_blocks(dev);
-
 	if (zbc_dev_is_zoned(dev)) {
-
 		/* Check RC_BASIS field */
 		switch ((cmd.buf[12] & 0x30) >> 4) {
-
 		case 0x00:
 			/*
 			 * The capacity represents only the space used by
@@ -649,26 +653,32 @@ static int zbc_scsi_get_capacity(struct zbc_device *dev)
 			 * To get the entire device capacity, we need to get
 			 * the last LBA of the last zone of the device.
 			 */
-			ret = zbc_scsi_do_report_zones(dev, 0, ZBC_RO_ALL|ZBC_RO_PARTIAL,
-						       &max_lba,
-						       NULL, &nr_zones);
+			zbc_debug("%s: READ CAPACITY RC_BASIS field is 0x00 "
+				  "(conventional zones capacity)\n",
+				  dev->zbd_filename);
+			ret = zbc_scsi_do_report_zones(dev, 0,
+						ZBC_RO_ALL | ZBC_RO_PARTIAL,
+						&max_lba, NULL, &nr_zones,
+						NULL,
+						dev->zbd_report_bufsz_min);
 			if (ret != 0)
 				goto out;
 
 			break;
-
 		case 0x01:
 			/* The device max LBA was reported */
-			break;
-
-		default:
-			zbc_error("%s: invalid RC_BASIS field encountered in READ CAPACITY result\n",
+			zbc_debug("%s: READ CAPACITY RC_BASIS field is 0x01 "
+				  "(all capacity)\n",
 				  dev->zbd_filename);
+			break;
+		default:
+			zbc_error("%s: invalid RC_BASIS field 0x%02x "
+				  "in READ CAPACITY result\n",
+				  dev->zbd_filename,
+				  (cmd.buf[12] & 0x30) >> 4);
 			ret = -EIO;
 			goto out;
-
 		}
-
 	}
 
 	/* Set the drive capacity using the reported max LBA */
@@ -710,9 +720,9 @@ int zbc_scsi_get_zbd_characteristics(struct zbc_device *dev)
 	if (!zbc_dev_is_zoned(dev))
 		return -ENXIO;
 
-	ret = zbc_scsi_inquiry(dev, 0xB6, buf, ZBC_SCSI_VPD_PAGE_B6_LEN);
+	ret = zbc_scsi_vpd_inquiry(dev, 0xB6, buf, ZBC_SCSI_VPD_PAGE_B6_LEN);
 	if (ret != 0) {
-		zbc_error("%s: zbc_scsi_inquiry VPD page 0xB6 failed\n",
+		zbc_error("%s: zbc_scsi_vpd_inquiry VPD page 0xB6 failed\n",
 			  dev->zbd_filename);
 		return ret;
 	}
@@ -772,6 +782,13 @@ static int zbc_scsi_get_dev_info(struct zbc_device *dev)
 {
 	int ret;
 
+	/*
+	 * Always request 512B aligned zone reports for in-kernel
+	 * ATA translation to work correctly.
+	 */
+	dev->zbd_report_bufsz_min = 512;
+	dev->zbd_report_bufsz_mask = dev->zbd_report_bufsz_min - 1;
+
 	/* Make sure the device is ready */
 	ret = zbc_sg_test_unit_ready(dev);
 	if (ret != 0)
@@ -792,6 +809,9 @@ static int zbc_scsi_get_dev_info(struct zbc_device *dev)
 	if (ret != 0)
 		return ret;
 
+	/* Get maximum command size */
+	zbc_sg_get_max_cmd_blocks(dev);
+
 	return 0;
 }
 
@@ -857,7 +877,7 @@ static int zbc_scsi_open(const char *filename,
 
 	*pdev = dev;
 
-	zbc_debug("%s: ########## SCSI driver succeeded ##########\n",
+	zbc_debug("%s: ########## SCSI driver succeeded ##########\n\n",
 		  filename);
 
 	return 0;
@@ -872,7 +892,7 @@ out:
 	if (fd >= 0)
 		close(fd);
 
-	zbc_debug("%s: ########## SCSI driver failed %d ##########\n",
+	zbc_debug("%s: ########## SCSI driver failed %d ##########\n\n",
 		  filename,
 		  ret);
 
